;===============================================================================
;
;	scboot.s - Steel Cap Boot
;
;	A file system independent, dual use Master Boot Record/Boot sector. 
;	Uses the partition table when booting as a MBR without a valid boot
;	loader sector. Allocates space for a BIOS Parameter Block, supports 
;	FAT12/16/32 and NT BPB. Includes support for 48 bit partition tables 
;	(uses C/H/S data area).

;	Requirements:
;	The device must include a valid boot loader sector when not being used
;	as a MBR.
;	The BPB must include valid values for headcnt, sectsz and trksectcnt
;	for C/H/S BIOS disk reads.
;
;===============================================================================
;
;	Released as Public Domain Software.
;
;	No Warranties expressed or implied.
;
;===============================================================================
;
;	0.0.1		2010-05-27 (bz) first release
;
;===============================================================================

%include "boot.inc"

	section	.text	
	bits	16		;8086 code only in the boot sector

	global	_start	
_start:
	jmp	short begin
	nop

	;BIOS Parameter Block	
bpb	istruc	fat16
	;use OEM name as data area
diskno:		db	0
chssect:	dw	0
chshead:	dw	0
at	fat16.sectsz,		dw	512
at	fat16.blksectcnt,	db	1
at	fat16.trksectcnt,	dw	18
at	fat16.headcnt,		dw	2
at	fat16.volsectcnt,	dw	2880
iend

	ORG(BPB_FAT32_END)

begin:
	;set segment bases
	cli
	xor	ax,ax
	mov	ds,ax
	mov	es,ax

	;create stack 
	mov	ss,ax			
	mov	sp,0x7c00
	sti

	;relocate to 0x0600 
	cld	
	mov	di,0x0600
	mov	si,0x7c00
	mov	cx,256
	rep	movsw
	jmp	0x0000:reloc
	
reloc:
	mov	[diskno],dl		;save BIOS disk number
	
bootloader:
	;check for boot loader address 
	mov	cx,4
	mov	si,loadsect
	xor	ax,ax
.1:
	cmpsw
	jnz	readloader
	loop	.1
	
	;check if on a HDD with a partition table
	test	dl,0x80
	jz	booterr			;not a HDD & no valid boot loader sector

chkparttbl:
	;find 1st active partion
	mov	si,parttbl
	mov	cx,4
.1:
	test	[si],byte 0x80		;test active bit
	jnz	partfound
	add	si,16
	loop	.1
	jmp	booterr
	
partfound:
	test	[si],byte 0x01		;check for 48 bit entry
	jz	.1
	;set high 16 bits of sector
	mov	ax,[si+0x02]
	mov	[loadsect+0x08],ax
.1:
	mov	ax,[si+0x08]
	mov	[loadsect],ax
	mov	ax,[si+0x0a]
	mov	[loadsect+0x04],ax
	
	;read boot sector
	mov	[extreadcnt],word 1
	mov	[extreadoff],word 0x7c00
	mov	[extreadseg],word 0
readloader:
	call	readsector
	push	ds
	push	word [extreadseg]
	pop	ds
	cmp	[0x7dfe],word 0xaa55	;boot signature check
	pop	ds
	jnz	booterr
	jmp	far word [extreadoff]

	;show error message and exit
booterr:
	mov	si,booterrmsg
	jmp	error	
diskerr:	
	mov	si,diskerrmsg
error:
	call	print
	mov	si,errmsg
	call	print
	jmp	$

;	BIOS print
;	
;
;entry:
;	DS:SI = string to print
;
_pc:
	mov	ah,0x0e
	mov	bx,0x0007
	int	0x10
print:
	lodsb
	cmp	al,0
	jnz	_pc
	ret

;	read a sector using BIOS calls
;
;on entry:
;	[extreadcnt] = number of sectors
;	[extreadbuf] = load address
;	[loadsect] = load sector offset
;
readsector:
	;check for BIOS disk extensions
	mov	dl,[diskno]
	test	dl,0x80
	jz	.chsread			;skip if floppy disk
	mov	ah,0x41
	mov	bx,0x55aa
	int	0x13
	jc	.chsread
	cmp	bx,0xaa55
	jnz	.chsread

.lbaread:
	;LBA sectors read
	mov	ah,42h
	mov	si,extread
	int	0x13
	jc	.chsread
	ret
	
.chsread:
	;Cylinder/Head/Sector read
	push	es
	mov	ax,[loadsect]
	mov	dx,[loadsect+2]
	mov	cx,[extreadcnt]
	mov	bx,[extreadseg]
	mov	es,bx
	mov	bx,[extreadoff]

.1:
	push	ax
	push	dx
	push	cx
	
	;convert a LBA sector to a CHS value
	cmp	dx,word [bpb+fat16.trksectcnt]	;overflow check
	ja	diskerr
	div	word [bpb+fat16.trksectcnt] 	;lba/sectors per track
	inc	dx
	mov	[chssect],dl			;save sect+1
	xor	dx,dx
	div	word [bpb+fat16.headcnt] 	;lba/heads (ax=cylinder)
	mov	[chshead],dl			;head (lba%heads)

	;read sector 
	mov	dx,ax
	mov	cl,6
	shl	dh,cl
	or	dh,[chshead]			;cylinder:2 bits/head:6 bits
	mov	dl,[diskno]			;disk no			
	mov	ch,al				;cylinder
	mov	cl,[chssect]			;sector
	mov	ax,0x0201
	int	0x13
	jc	diskerr
	
	pop	cx
	pop	dx
	pop	ax
	
	inc	ax					;next sector
	adc	dx,0
	add	bx,[bpb+fat16.sectsz]			;adjust memory offset
	loop	.1
	pop	es
	ret
	
booterrmsg:	db	"Boot",0
diskerrmsg:	db	"Disk",0
errmsg:		db	" error",0
	
	ORG(EXTREAD_ADDR)
	;BIOS extended disk read
extread:	db	0x10, 0x00
extreadcnt:	dw	0x0008		;load 4kb for boot loader
extreadoff:	dw	0x7c00		
extreadseg:	dw	0x0000		;load to 0x0000:0x7c00
loadsect:	dd	1,0		;boot loader lba sector

	ORG(PART_TBL)
parttbl:
	
	ORG(BOOT_SIG)
	dw	0xaa55
